/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.controllers.voxels;

import fr.amap.commons.util.swing.jprogress.JProgressEvent;
import fr.amap.commons.util.swing.jprogress.JProgressListener;
import fr.amap.commons.util.swing.jprogress.JProgressable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * Voxel space file parsing class.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class VoxelSpaceFile extends JProgressable {
    
    private final List<String> names;
    private final Map<String, Float> max;
    private final Map<String, Float> min;
    private final List<List<String>> lines;
    private int width, height, depth;
    
    public VoxelSpaceFile() {
        super();
        this.names = new ArrayList();
        this.lines = new ArrayList();
        this.max = new HashMap();
        this.min = new HashMap();
        this.width = 0;
        this.height = 0;
        this.depth = 0;
    }
    
    public void setSize(int width, int height, int depth) {
        this.width = width;
        this.height = height;
        this.depth = depth;
    }
    
    public int getWidth() {
        return width;
    }
    
    public int getHeight() {
        return height;
    }
    
    public int getDepth() {
        return depth;
    }
    
    public final void addName(final String name) {
        names.add(name);
    }
    
    public final void addLine(final List<String> line) {
        lines.add(line);
    }
    
    public String[] getNames() {
        String[] array = new String[names.size()];
        
        for (int i = 0; i < array.length; i++) {
            array[i] = names.get(i);
        }
        
        return array;
    }
    
    public final String getProperty(String name, int index) {
        if (lines.size() > index) {
            if (names.contains(name)) {
                return lines.get(index).get(names.indexOf(name));
            }
        }
        
        return null;
    }
    
    public final float getMinPropertyValue(String name) {
        if (min.isEmpty()) {
            throw(new NullPointerException(
                    "The " + name + " property doesn't contains any values."));
        }
        
        return min.get(name);
    }
    
    public final float getMaxPropertyValue(String name) {    
        if (max.isEmpty()) {
            throw(new NullPointerException(
                    "The " + name + " property doesn't contains any values."));
        }
        
        return max.get(name);
    }
    
    public final List<String> getLine(String name) {
        List<String> values = new ArrayList<>();
        
        for (int i = 0; i < lines.size(); i++) {
            values.add(getProperty(name, i));
        }
        
        return values;
    }
    
    public boolean parse(final InputStream voxelFile) {                      
        if (voxelFile != null) {
            names.clear();
            min.clear();
            max.clear();
            lines.clear();
            
            try {
                final BufferedReader br = new BufferedReader(new InputStreamReader(voxelFile));
                                
                if (br.ready()) {
                    for (JProgressListener progressListener : getProgressListeners()) {
                        progressListener.onProgress(new JProgressEvent(
                                1, 3, 
                                "Check if it's a Voxel Space file..."));
                    }
                    
                    if (br.readLine().contains("VOXEL SPACE")) {                        
                        String line;
                        
                        for (JProgressListener progressListener : getProgressListeners()) {
                            progressListener.onProgress(new JProgressEvent(
                                    2, 3, 
                                    "Start parsing file..."));
                        }
                        
                        while ((line = br.readLine()) != null) {                            
                            if (!line.startsWith("#") && !line.startsWith("VOXEL SPACE")) {
                                if (line.startsWith("i j k")) {
                                    for (String name : line.split(" ")) {
                                        addName(name);
                                    }
                                } else {
                                    List<String> values = new ArrayList();
                                    
                                    for (String value : line.split(" ")) {
                                        values.add(value);
                                    }
                                    
                                    addLine(values);
                                }
                            } else if (line.startsWith("#split")) {
                                String[] values = line.substring(8).split(" ");
                                
                                setSize(Integer.parseInt(values[0]),
                                        Integer.parseInt(values[1]),
                                        Integer.parseInt(values[2]));
                            }
                        }
                        
                        for (JProgressListener progressListener : getProgressListeners()) {
                            progressListener.onProgress(new JProgressEvent(
                                    3, 3, 
                                    "Set min and max values..."));
                        }
                        
                        names.forEach((name) -> {
                            for (int i = 0; i < lines.size(); i++) {
                                float value = Float.valueOf(getProperty(name, i));

                                if (!Float.isNaN(value)) {
                                    if (!max.containsKey(name)) {
                                        max.put(name, value);
                                    } else if (value > max.get(name)) {
                                        max.put(name, value);
                                    }
                                    
                                    if (!min.containsKey(name)) {
                                        min.put(name, value);
                                    } else if (value < min.get(name)) {
                                        min.put(name, value);
                                    }
                                }
                            }
                        });
                        
                        if (max.isEmpty() || min.isEmpty()) {
                            return false;
                        }
                    } else {
                        Logger.getLogger(VoxelSpaceFile.class.getName())
                                .log(Level.ERROR, "The voxel space file doesn't "
                                + "have a 'Voxel Space' header.");
                        return false;
                    }
                    
                    br.close();
               }
            } catch (IOException ex) {
                ex.printStackTrace();
                Logger.getLogger(VoxelSpaceFile.class.getName()).log(Level.ERROR, null, ex);
                return false;
            }
        } else {
            Logger.getLogger(VoxelSpaceFile.class.getName()).log(Level.ERROR, "The voxel space file is null.");
            return false;
        }
        
        return true;
    }
    
}
